import sys
import click


class CobException(Exception):
    pass


class CobExecutionError(CobException, click.ClickException):
    exit_code = -1

    def show(self, file=None):
        if file is not None:
            file = sys.stderr
        click.echo(click.style(f'Error: {self.format_message()}', fg='red'), file=file)


class NotInProject(CobExecutionError):
    pass


class CobConflict(CobExecutionError):
    pass


class BadMountpoint(CobConflict):
    pass


class MountpointConflict(CobConflict):
    pass


class MissingDependency(CobExecutionError):
    pass


class UnknownSubsystem(CobExecutionError):
    pass


class TestsFailed(CobExecutionError):
    pass



import logbook
from pathlib import Path
import pytest
from uuid import uuid4
import yaml

from .project import Project


config = None


def pytest_addoption(parser):
    parser.addoption("--prebuilt", action="store_true", default=False,
        help="Use prebuilt docker images")


@pytest.fixture(autouse=True)
def configure(request):
    global config # pylint: disable=global-statement
    config = request.config

    @request.addfinalizer
    def cleanup():    # pylint: disable=unused-variable
        global config # pylint: disable=trailing-whitespace, global-statement
        config = None



@pytest.fixture(scope='session', autouse=True)
def setup_logging():
    logbook.StderrHandler(level=logbook.TRACE).push_application()


@pytest.fixture
def project_name():
    return f'proj{str(uuid4()).replace("-", "_")}'


@pytest.fixture
def empty_project(tmpdir, project_name):
    returned = Project(tmpdir, project_name)
    returned.generate_project()
    return returned


class _ProjectFactory:

    def __init__(self, root):
        self.root = root
        self.config = {}

    def build(self):
        from cob.project import Project
        with (self.root / '.cob-project.yml').open('w') as f:
            yaml.dump(self.config, f)
        return Project(self.root)


@pytest.fixture
def project_factory(tmpdir):
    tmpdir = Path(tmpdir)
    return _ProjectFactory(tmpdir)



import os
import click
import logbook
import multiprocessing
from pathlib import Path
import random
import shutil
import string
import subprocess
import sys
from tempfile import mkdtemp

import gunicorn.app.base
from werkzeug.contrib.fixers import ProxyFix
import yaml


from ..app import build_app
from ..bootstrapping import ensure_project_bootstrapped
from ..exceptions import TestsFailed
from ..utils.config import get_etc_config_path
from ..utils.docker import docker_cmd, docker_compose_cmd
from ..utils.develop import is_develop, cob_root
from ..utils.network import wait_for_app_services, wait_for_tcp
from ..utils.templates import load_template
from ..project import get_project

import pkg_resources


_COB_VERSION = pkg_resources.get_distribution('cob').version  # pylint: disable=no-member

_logger = logbook.Logger(__name__)
_CUSTOM_DOCKERFILE = "custom.docker"

def _get_user_steps():
    if not os.path.isfile(_CUSTOM_DOCKERFILE):
        return ''

    with open(_CUSTOM_DOCKERFILE) as f:
        return f.read()


@click.group()
def docker():
    pass


@docker.command(name="generate-docker-file")
def generate_dockerfile():
    proj = get_project()
    template = load_template('Dockerfile')

    if is_develop():
        sdist_file_name = os.path.join(proj.root, '.cob-sdist.tar.gz')
        _build_cob_sdist(filename=sdist_file_name)
    else:
        sdist_file_name = None

    with open(".Dockerfile", "w") as f:
        f.write(template.render(
            project=proj,
            deployment_base_image='python:3.6-jessie',
            python_version='3.6',
            is_develop=is_develop(),
            cob_sdist_filename=os.path.basename(sdist_file_name) if sdist_file_name else None,
            cob_root=cob_root() if is_develop() else None,
            user_steps=_get_user_steps()))


def _build_cob_sdist(filename):

    sdist_filename = os.environ.get('COB_SDIST_FILENAME')

    if sdist_filename is not None:
        shutil.copy(sdist_filename, filename)
    else:
        tmpdir = mkdtemp()
        try:
            subprocess.check_call(
                f'python setup.py sdist -d {tmpdir}', cwd=cob_root(), shell=True)
            [distfile] = os.listdir(tmpdir)
            sdist_filename = os.path.join(tmpdir, distfile)
            shutil.move(sdist_filename, str(filename))
        finally:
            shutil.rmtree(tmpdir)


@docker.command(name='build')
@click.option('--sudo/--no-sudo', is_flag=True, default=None, help="Run docker build with sudo")
@click.option('--extra-build-args', '-e', default="", help="Arguments to pass to docker build")
@click.option('--release', is_flag=True, default=False)
def docker_build(sudo, extra_build_args="", use_exec=True, image_name=None, release=False):
    project = get_project()
    if image_name is None:
        image_name = f'{project.get_docker_image_name()}:{"latest" if release else "dev"}'.format(project.name, 'latest' if release else 'dev')
    generate_dockerfile.callback()
    cmd = docker_cmd.build(['-t', image_name, '-f', '.Dockerfile', '.', *extra_build_args.split()]).force_sudo(sudo)
    _logger.debug('Running Command: {}', cmd)
    cmd.run(use_exec=use_exec)


@docker.command(name='wsgi-start')
def start_wsgi():
    logbook.StderrHandler(level=logbook.DEBUG).push_application()

    _ensure_secret_config()
    ensure_project_bootstrapped()
    project = get_project()
    app = build_app(config_overrides={'PROPAGATE_EXCEPTIONS': True})
    app.wsgi_app = ProxyFix(app.wsgi_app)

    wait_for_app_services(app)

    if project.subsystems.has_database():
        with app.app_context():
            project.setup_db()

    workers_count = (multiprocessing.cpu_count() * 2) + 1

    class StandaloneApplication(gunicorn.app.base.BaseApplication):  # pylint: disable=abstract-method

        def __init__(self, app, options=None):
            self.options = options or {}
            self.application = app
            super(StandaloneApplication, self).__init__()

        def load_config(self):
            config = {key: value for key, value in self.options.items()
                      if key in self.cfg.settings and value is not None}

            for key, value in config.items():
                self.cfg.set(key.lower(), value)

        def load(self):
            return self.application

    options = {
        'bind': '0.0.0.0:8000',
        'workers': workers_count,
    }
    options.update(project.config.get('gunicorn', {}))

    StandaloneApplication(app, options).run()


def _ensure_secret_config():
    conf_dir = os.environ.get('COB_CONFIG_DIR')
    if not conf_dir:
        return
    secret_file = os.path.join(conf_dir, '000-cob-private.yml')
    if os.path.isfile(secret_file):
        return

    with open(secret_file, 'w') as f:
        f.write('flask_config:\n')
        for secret_name in ('SECRET_KEY', 'SECURITY_PASSWORD_SALT'):
            f.write(f'  {secret_name}: {_generate_secret_string()!r}\n')


def _generate_secret_string(length=50):
    return "".join([random.choice(string.ascii_letters) for i in range(length)])


@docker.command(name='nginx-start')
@click.option('--print-config', is_flag=True, default=False)
def start_nginx(print_config):
    project = get_project()
    template = load_template('nginx_config')
    config = template.render({'use_ssl': False, 'hostname': None, 'project': project, 'os': os})

    if print_config:
        print(config)
        return

    wait_for_tcp('wsgi', 8000)

    with open('/etc/nginx/conf.d/webapp.conf', 'w') as f:
        f.write(config)

    nginx_path = '/usr/sbin/nginx'
    os.execv(nginx_path, [nginx_path, '-g', 'daemon off; error_log /dev/stdout info;'])


@docker.command()
@click.option('--http-port', default=None)
@click.option('--build', is_flag=True, default=False)
@click.option('compose_overrides', '-o', '--overlay-compose-file', default=[], multiple=True)
@click.option('-d', '--detach', is_flag=True, default=False)
@click.option('--image-name', default=None, help='Image to use for the main docker containers')
def run(http_port, build, detach, image_name, compose_overrides):
    project = get_project()
    if build:
        docker_build.callback(sudo=False, use_exec=False)
    if image_name is None:
        image_name = f'{project.get_docker_image_name()}:dev'
    run_image.callback(image_name=image_name, detach=detach, http_port=http_port, compose_overrides=compose_overrides)


@docker.command()
def stop():
    _exec_docker_compose(['down'])


@docker.command()
def logs():
    _exec_docker_compose(['logs'])


def _exec_docker_compose(cmd, **kwargs):
    project = get_project()
    compose_filename = f'/tmp/__{project.name}-docker-compose.yml'
    with open(compose_filename, 'w') as f:
        f.write(_generate_compose_file_string(**kwargs))
    cmd = docker_compose_cmd.args(['-f', compose_filename, '-p', project.name, *cmd])
    cmd.execv()


@docker.command(name='generate-docker-compose-file')
@click.option('--image-name', default=None)
@click.option('--http-port', type=int, default=None)
@click.option('--force-config-override', is_flag=True, default=False)
@click.option('--log-driver', default='syslog')
def generate_docker_compose_file(image_name, force_config_override, http_port, log_driver):
    """Prints out a docker-compose.yml for this project"""
    print(_generate_compose_file_string(image_name=image_name, force_config_override=force_config_override, http_port=http_port, log_driver=log_driver))


def _generate_compose_file_dict(*, http_port=None, image_name=None, force_config_override=False, log_driver='syslog'):
    project = get_project()

    if image_name is None:
        image_name = f'{project.get_docker_image_name()}:dev'

    config = {
        'version': '3',
        'volumes': {
            'conf': None
        },
    }

    common_environment = {
        'COB_DATABASE_URI': 'postgresql://{0}@db/{0}'.format(project.name),
        'COB_CELERY_BROKER_URL': 'amqp://guest:guest@rabbitmq',
        'COB_CONFIG_DIR': '/conf',
    }

    services = config['services'] = {

        'wsgi':  {
            'image': image_name,
            'command': 'cob docker wsgi-start',
            'environment': common_environment,
            'depends_on': [],
            'volumes': [
                'conf:/conf',
            ],
        },

        'nginx': {
            'image': image_name,
            'command': 'cob docker nginx-start',
            'ports': [f'{http_port or 8000}:80'],
            'depends_on': ['wsgi'],
        }
    }

    if project.subsystems.has_database():
        services['wsgi']['depends_on'].append('db')
        services['db'] = {
            'image': 'postgres:9.6',
            'volumes': [
                'db:/var/lib/postgresql/data'
            ],
            'environment': {
                'POSTGRES_USER': project.name,
                'POSTGRES_DB': project.name,
            }
        }
        config['volumes']['db'] = None

    if project.subsystems.has_tasks():

        services['rabbitmq'] = {
            'image': 'rabbitmq',
        }
        services['worker'] = {
            'image': image_name,
            'command': 'cob celery start-worker',
            'environment': common_environment,
        }

    if project.services.redis:
        services['redis'] = {
            'image': 'redis',
        }

    config_override_path = get_etc_config_path(project.name).resolve()
    if force_config_override or config_override_path.is_dir():
        for service_config in services.values():
            service_config.setdefault('volumes', []).append('{0}:{0}'.format(config_override_path))

    for service_config in services.values():
        service_config.setdefault('volumes', []).append(
            '/etc/localtime:/etc/localtime:ro',
        )

        service_config.setdefault('logging', {'driver': log_driver})

    for service_name, service_ports in project.config.get('docker', {}).get('exposed_ports', {}).items():
        services[service_name].setdefault('ports', []).extend(service_ports)

    return config


def _generate_compose_file_string(**kwargs):
    config = _generate_compose_file_dict(**kwargs)
    return _dump_yaml(config)


def _dump_yaml(config, *, stream=None):
    return yaml.safe_dump(config, stream, allow_unicode=True, default_flow_style=False)


@docker.command()
@click.option('build_image', '--no-build', is_flag=True, default=True)
@click.option('--sudo/--no-sudo', is_flag=True, default=None, help="Run docker build with sudo")
def test(build_image, sudo):
    project = get_project()
    image_name = f"{project.get_docker_image_name()}:dev"
    if build_image:
        docker_build.callback(sudo=sudo, use_exec=False, image_name=image_name)
    compose_file_dict = _generate_compose_file_dict(image_name=image_name)
    compose_file_dict['services'].pop('nginx')
    test_config = compose_file_dict['services'].pop('wsgi')

    test_config['tty'] = True
    test_config['depends_on'] = sorted(set(compose_file_dict['services']) - {'test'})
    test_config['stdin_open'] = True
    compose_file_dict['services']['test'] = test_config

    compose_filename = f'/tmp/__{project.name}-test-docker-compose.yml'
    with open(compose_filename, 'w') as f:
        _dump_yaml(compose_file_dict, stream=f)
    docker_compose_name = f'{project.name}-test'
    cmd = docker_compose_cmd.args([
        '-f', compose_filename, '-p', docker_compose_name, 'run',
        '-w', '/app', '-v', f'{os.path.abspath(".")}:/localdir',
        'test',
        'bash', '-c', "rsync -rvP --delete --exclude .cob /localdir/ /app/ && cob test --migrate"])
    if cmd.popen().wait() != 0:
        raise TestsFailed('Tests failed')


@docker.command(name="run-image", help='Runs a cob project in a pre-built docker image')
@click.argument('image_name')
@click.option('-d', '--detach', is_flag=True, default=False)
@click.option('compose_overrides', '-o', '--overlay-compose-file', default=[], multiple=True)
@click.option('--http-port', default=None)
def run_image(image_name, detach, compose_overrides, http_port):
    project_name, compose_path = _generate_compose_file_from_image(image_name, http_port=http_port)
    cmd = docker_compose_cmd.args(['-p', project_name, '-f', compose_path])
    for compose_override in compose_overrides:
        cmd.args(['-f', compose_override])
    cmd.args(['up'])
    if detach:
        cmd = cmd.args(['-d'])

    cmd.execv()


@docker.command(name="stop-image")
@click.argument('image_name')
def stop_image(image_name):
    project_name, compose_path = _generate_compose_file_from_image(image_name)
    docker_compose_cmd.args(['-p', project_name, '-f', compose_path, 'down']).execv()


def _generate_compose_file_from_image(image_name, *, http_port=None):
    project_name = _get_project_name_from_image(image_name)
    cmd = docker_cmd.run(['--rm', image_name, 'cob', 'docker', 'generate-docker-compose-file', '--image-name', image_name, '--http-port', '80'])
    if _is_journald_system():
        cmd.args(['--log-driver=journald'])
    if get_etc_config_path(project_name).is_dir():
        cmd.args(['--force-config-override'])
    if http_port is not None:
        cmd.args(['--http-port', str(http_port)])
    compose_file_contents = cmd.check_output()
    compose_path = Path('/tmp') / f"__cob_docker_compose_{image_name.replace(':', '_').replace('/', '_')}.yml"
    with compose_path.open('wb') as f:
        f.write(compose_file_contents)
    return project_name, compose_path


@docker.command(name='deploy', help='Deploys an dockerized cob app image to the local systemd-based machine')
@click.option('--force', is_flag=True, default=False)
@click.argument('image_name')
def deploy_image(image_name, force):
    click.echo(f'Obtaining project information for {image_name}...')
    project_name = _get_project_name_from_image(image_name)
    unit_template = load_template('systemd_unit')
    filename = Path('/etc/systemd/system') / f'{project_name}-docker.service'
    click.echo(f'Writing systemd unit file under {filename}...')
    if filename.exists() and not force:
        click.confirm(f'{filename} already exists. Overwrite?', abort=True)

    tmp_filename = Path(mkdtemp()) / 'systemd-unit'
    with tmp_filename.open('w') as f: # pylint: disable=no-member
        f.write(unit_template.render(project_name=project_name, image_name=image_name, cob=f'{sys.executable} -m cob.cli.main'))

    subprocess.check_call(f'sudo -p "Enter password to deploy service" mv {tmp_filename} {filename}', shell=True)
    click.echo(f'Starting systemd service {filename.stem}...')
    subprocess.check_call('sudo systemctl daemon-reload', shell=True)
    subprocess.check_call(f'sudo systemctl enable --now {filename.name}', shell=True)


def _get_project_name_from_image(image_name):
    return docker_cmd.run(['--rm', image_name, 'cob', 'info', 'project-name']).check_output(encoding='utf-8').strip()


def _is_journald_system():
    return shutil.which('journalctl') is not None


@docker.command(name='tag-latest', help='Tags the latest development image as latest')
def tag_latest():
    project = get_project()
    docker_cmd.tag([
        f'{project.get_docker_image_name()}:dev',
        f'{project.get_docker_image_name()}:latest']
    ).execv()

@docker.command(name='push')
def push():
    project = get_project()
    docker_cmd.push([f'{project.get_docker_image_name()}:latest']).execv()



from mitba import cached_function
from ..exceptions import MissingDependency
import os
import shutil
import subprocess


__all__ = ['docker_cmd', 'docker_compose_cmd']


class Docker:

    def __init__(self, command_name='docker'):
        self._cmdname = command_name
        self._cmd_full_path = None

    def get_command_name(self):
        return self._cmdname

    @cached_function
    def is_sudo_needed(self):
        proc = subprocess.Popen(
            'docker ps', shell=True, stderr=subprocess.PIPE, stdout=subprocess.DEVNULL)
        err = proc.communicate()[1].decode('utf-8')
        if proc.returncode == 0:
            return False
        assert 'permission denied' in err.lower(), f"'docker ps' failed on {err}"
        return True

    def get_full_command_name(self):
        if self._cmd_full_path is None:
            cmd = self._cmdname
            if not os.path.isabs(cmd):
                cmd = shutil.which(cmd)
                if cmd is None:
                    raise MissingDependency(f'{self._cmdname} could not be found')
                self._cmd_full_path = cmd
        return self._cmd_full_path

    def args(self, args):
        return DockerCommand(self, args)

    def __getattr__(self, attr):
        if attr.startswith('_'):
            raise AttributeError(attr)

        def returned(args):
            return self.args([attr]).args(args)

        returned.__name__ = attr
        return returned


class DockerCommand:

    def __init__(self, docker, args):
        self._docker = docker
        self._force_sudo = False
        self._args = list(args)

    def args(self, args):
        self._args.extend(args)
        return self

    def force_sudo(self, force=True):
        self._force_sudo = force
        return self

    def to_split_command(self):
        returned = []
        if self._force_sudo or self._docker.is_sudo_needed():
            returned.extend(
                ['sudo', '-p', 'Please enter your password to run docker: '])
        returned.append(self._docker.get_full_command_name())
        returned.extend(self._args)
        return returned

    def execv(self):
        cmd = self.to_split_command()
        os.execv(cmd[0], cmd)

    def popen(self, *args, **kwargs):
        return subprocess.Popen(self.to_split_command(), *args, **kwargs)

    def check_output(self, *args, **kwargs):
        return subprocess.check_output(self.to_split_command(), *args, **kwargs)

    def run(self, use_exec=False):
        if use_exec:
            self.execv()
        else:
            returned = self.popen()
            if returned.wait() != 0:
                raise subprocess.CalledProcessError(
                    returned.returncode, self.to_split_command())
            return returned

    def __repr__(self):
        return f"[docker] {' '.join(self._args)}"


docker_cmd = Docker()
docker_compose_cmd = Docker('docker-compose')



# cob: mountpoint=/
from cob import route
from cob import services

@route('/test_redis')
def test_redis():
    services.redis.set('key', 'value')
    value = services.redis.get('key')
    assert value.decode('utf-8') == 'value', f'Got {value!r} instead!'
    return 'ok'



import click
from halo import Halo
import logbook
import pathlib
import subprocess
import os
import sys
import tempfile
import yaml


_logger = logbook.Logger(__name__)

_ROOT = pathlib.Path(__file__).parent.parent / '_test_projects'



@click.command()
def main():
    from cob.cli.docker_cli import _build_cob_sdist

    tmpdir = pathlib.Path(tempfile.mkdtemp())
    sdist_file = tmpdir / 'cob_sdist.tar.gz'
    logbook.StderrHandler(level=logbook.DEBUG).push_application()

    _build_cob_sdist(sdist_file)
    os.environ['COB_SDIST_FILENAME'] = str(sdist_file)

    projects = []
    processes = []
    logs = []

    project_names = set()

    for project_dir in _ROOT.iterdir():
        proj_yaml = project_dir / '.cob-project.yml'
        if proj_yaml.exists():
            with proj_yaml.open() as f:
                name = yaml.load(f.read()).get('name')
                if name is None:
                    sys.exit(f'Project {project_dir.stem} does not have a name configured')
                if name in project_names:
                    sys.exit(f'Project name {name} is already in use (found in {project_dir})')
                project_names.add(name)

            projects.append(proj_yaml)


    for proj_yaml in projects:
        project_dir = proj_yaml.parent
        log = tmpdir / project_dir.stem
        logs.append(log)
        with log.open('w') as logfile:
            _logger.info('Building {}...', project_dir)
            processes.append(
                subprocess.Popen(f'{sys.executable} -m cob.cli.main docker build', cwd=project_dir, stderr=subprocess.STDOUT, stdout=logfile, shell=True)
            )
    success = True

    with Halo('Waiting for projects to build'):
        for p, log in zip(processes, logs):
            project = log.stem
            result = p.wait()
            if result != 0:
                success = False
                click.echo(click.style(f'Build failed for {project}. More details below', fg='red'))
                with log.open() as f:
                    for line in f:
                        click.echo(click.style(line, fg='yellow'))
                click.echo(click.style('^' * 80, fg='red'))

    if not success:
        raise click.ClickException('Some build processes failed')


if __name__ == '__main__':
    main()



import redis

_DISABLED = {'enabled': False}

class Services:

    redis = None

    def __init__(self, project):

        service_config = project.config.get('services')
        if service_config is not None:
            if not isinstance(service_config, dict):
                service_config = {service_name: {'enabled': True} for service_name in service_config}
            service_config = service_config.copy()

            if service_config.pop('redis', _DISABLED).get('enabled', True):
                self.redis = redis.Redis('redis' if project.is_dockerized() else None)


class ServiceProxy:

    def __getattr__(self, attr):
        if not attr.startswith('_'):
            project = get_project()

            return getattr(project.services, attr)
        return super().__getattribute__(attr)


services = ServiceProxy()

from .project import get_project



import copy
import os
from pathlib import Path

import yaml


def get_etc_config_path(project_name):
    return Path('/etc/cob/conf.d') / project_name


def get_config_override_locations(project_name):
    return [
        get_etc_config_path(project_name),
        Path('~/.config/cob/projects').expanduser() / project_name,
    ]


def load_overrides(root, *, environ=None):

    if environ is None:
        environ = os.environ

    config_dirs = []
    project_name = root.get('name', None)
    if project_name is not None:
        config_dirs.extend(get_config_override_locations(project_name))
    config_dir_env = environ.get('COB_CONFIG_DIR')
    if config_dir_env:
        config_dirs.append(Path(config_dir_env))

    for config_dir in config_dirs:
        if not isinstance(config_dir, Path):
            config_dir = Path(config_dir)

        if config_dir.is_dir():
            for yaml_file in sorted(config_dir.iterdir(), key=lambda p: p.stem):
                if yaml_file.stem.startswith(('_', '.')) or yaml_file.suffix != '.yml':
                    continue
                with yaml_file.open() as f:
                    root = merge_config(root, yaml.load(f))
    return root


def merge_config(parent, *children, in_place=False):
    if in_place:
        returned = parent
    else:
        returned = copy.deepcopy(parent)
    for child in children:
        for key, value in child.items():
            if isinstance(value, dict):
                if key not in returned:
                    returned[key] = value.copy()
                else:
                    if not isinstance(returned[key], dict):
                        raise ValueError(f'Cannot merge dictionaries: value of {key!r} is not a dictionary')
                    merge_config(returned[key], value, in_place=True)
            else:
                returned[key] = copy.deepcopy(value)
    return returned



#!/usr/bin/env python
import os
import pkg_resources
import sys

import logbook
import click

from cob.app import build_app
from cob.project import get_project
from cob.exceptions import CobExecutionError
from cob.bootstrapping import ensure_project_bootstrapped

@click.group()
@click.option("-v", "--verbose", count=True)
@click.option("-q", "--quiet", count=True)
def main(verbose, quiet):
    logbook.NullHandler().push_application()
    logbook.StreamHandler(sys.stderr, level=logbook.WARNING -
                          verbose + quiet, bubble=False).push_application()

@main.command()
def version():
    print(pkg_resources.get_distribution('cob').version)


@main.add_command
@click.command()
def bootstrap():
    try:
        ensure_project_bootstrapped()
    except RuntimeError as e:
        raise CobExecutionError(str(e))


@main.add_command
@click.command(context_settings={
    'ignore_unknown_options': True
})
@click.option('--migrate', is_flag=True, default=False)
@click.argument('pytest_args', nargs=-1, type=click.UNPROCESSED)
def test(pytest_args, migrate):
    ensure_project_bootstrapped()
    python_executable = os.path.abspath('.cob/env/bin/python')
    if migrate:
        with build_app().app_context():
            get_project().setup_db()
    os.execv(python_executable, [python_executable, '-m', 'pytest'] + list(pytest_args))


@main.add_command
@click.command()
def shell():
    ensure_project_bootstrapped()
    app = build_app()

    sys.path[:] = [p for p in sys.path if p not in ('', '.')]


    with app.app_context():
        ns = {'app': app, '__package__': '_cob'}
        try:
            from IPython import embed
        except ImportError:
            import code
            code.interact(local=ns)
        else:
            embed(user_ns=ns)



def _add_all_subcommands():
    for name in [
            'celery',
            'db',
            'develop',
            'docker',
            'generate',
            'info',
            'testserver',
            'list',
            'migrate',
    ]:
        mod = __import__(f'cob.cli.{name}_cli', fromlist=[''])
        cmd = getattr(mod, name)
        main.add_command(cmd)

_add_all_subcommands()


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



# cob: mountpoint=/config
from cob import route
from cob.project import get_project

from flask import jsonify, current_app


@route('/')
def get_config():
    return jsonify({
        'flask_config': {
            key: current_app.config[key] for key in
            ['SECRET_KEY',
             'SQLALCHEMY_TRACK_MODIFICATIONS']
        },
        'cob_config': get_project().config,
    })



import re

from .project import Project


def test_ember_apps():
    with Project('ember').server_context() as app:

        for asset_uri in _iter_assets(app):
            assert app.get(asset_uri)

        for uri in ['/', '/items']:
            assert 'vendor' in app.get(uri).text

def test_ember_non_root_uri():
    with Project('ember_non_root_uri').server_context() as app:
        assert app.get('/').text == 'root resource here'
        assert 'link rel="stylesheet"' in app.get('/ui/').text


def _iter_assets(app):
    page = app.get('/').text
    yield from re.findall(r'<link rel="stylesheet" href="([^"]*)">', page)
    yield from re.findall(r'<script src="([^"]*)">', page)



import os
import sys
import yaml
from pathlib import Path


_ROOT = Path(__file__).parent.parent
_TEST_ROOT = _ROOT / 'tests'
_TRAVIS_YML = _ROOT / '.travis.yml'


def main():
    worker_id = int(os.environ['WORKER'])
    assert worker_id > 0

    tests = sorted([test for test in _TEST_ROOT.iterdir() if test.stem.startswith('test_') and test.suffix == '.py'],
                   key = lambda test: test.stem)
    assert tests

    with _TRAVIS_YML.open() as f:
        config = yaml.load(f.read())

    num_workers = len([x for x in config['jobs']['include'] if str(x['script']).strip().startswith('WORKER=')])
    assert num_workers > 0

    tests = [str(test) for index, test in enumerate(tests) if index % num_workers == worker_id - 1]
    print('Running tests:', tests)

    argv = [sys.executable, '-m', 'pytest', *tests, *sys.argv[1:]]
    os.execv(argv[0], argv)

if __name__ == '__main__':
    main()



import click
from ..project import get_project


@click.group()
def info():
    pass

@info.command(name='project-name')
def project_name():
    """Prints out the name of the current project"""
    print(get_project().name)


@info.command(name='docker-image-name')
def docker_image_name():
    """Prints out the name of the docker image to use for this project"""
    print(get_project().get_docker_image_name())
